# SPDX-License-Identifier: Apache-2.0

import functools
import copy
import ipaddress
from threading import RLock

from celery import Celery
from celery.signals import worker_process_init
import jinja2
import keystoneauth1
import kombu.utils
import openstack
from pottery import Redlock
from redis import Redis

from osism import settings
from osism.tasks import Config, conductor, netbox
from osism import utils

# https://github.com/celery/kombu/issues/1804
if not getattr(kombu.utils.cached_property, "lock", None):
    setattr(
        kombu.utils.cached_property,
        "lock",
        functools.cached_property(lambda _: RLock()),
    )
    # Must call __set_name__ here since this cached property is not defined in the context of a class
    # Refer to https://docs.python.org/3/reference/datamodel.html#object.__set_name__
    kombu.utils.cached_property.lock.__set_name__(kombu.utils.cached_property, "lock")

app = Celery("openstack")
app.config_from_object(Config)

redis = None
conn = None


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global conn
    global redis

    redis = Redis(
        host=settings.REDIS_HOST,
        port=settings.REDIS_PORT,
        db=settings.REDIS_DB,
        socket_keepalive=True,
    )
    redis.ping()

    # Parameters come from the environment, OS_*
    try:
        conn = openstack.connect()
    except keystoneauth1.exceptions.auth_plugins.MissingRequiredOptions:
        pass


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    pass


@app.task(bind=True, name="osism.tasks.openstack.image_get")
def image_get(self, image_name):
    result = conn.image.find_image(image_name)
    return result.id


@app.task(bind=True, name="osism.tasks.openstack.network_get")
def network_get(self, network_name):
    result = conn.network.find_network(network_name)
    return result.id


@app.task(bind=True, name="osism.tasks.openstack.baremetal_node_show")
def baremetal_node_show(self, node_id_or_name):
    result = conn.baremetal.find_node(node_id_or_name)
    return result


@app.task(bind=True, name="osism.tasks.openstack.baremetal_node_list")
def baremetal_node_list(self):
    nodes = conn.baremetal.nodes()
    result = []

    # Simulate the output of the OpenStack CLI with -f json and without --long
    for node in nodes:
        result.append(
            {
                "UUID": node.id,
                "Name": node.name,
                "Instance UUID": node.instance_id,
                "Power State": node.power_state,
                "Provisioning State": node.provision_state,
                "Maintenance": node.is_maintenance,
            }
        )

    return result


@app.task(
    bind=True, name="osism.tasks.openstack.baremetal_introspection_interface_list"
)
def baremetal_introspection_interface_list(self, node_id_or_name):
    pass


@app.task(bind=True, name="osism.tasks.openstack.baremetal_introspection_status")
def baremetal_introspection_status(self, node_id_or_name):
    result = None
    return result


@app.task(bind=True, name="osism.tasks.openstack.baremetal_get_network_interface_name")
def baremetal_get_network_interface_name(self, node_name, mac_address):
    global conn

    introspection = conn.baremetal_introspection.get_introspection(node_name)

    # Wait up to 5 minutes for the completion of a running introspection
    conn.baremetal_introspection.wait_for_introspection(introspection, timeout=30)

    introspection_data = conn.baremetal_introspection.get_introspection_data(
        introspection
    )
    interfaces = introspection_data["inventory"]["interfaces"]

    result = None
    for interface in interfaces:
        if interface["mac_address"].lower() == mac_address.lower():
            result = interface["name"]

    return result


@app.task(bind=True, name="osism.tasks.openstack.baremetal_set_node_provision_state")
def baremetal_set_node_provision_state(self, node, state):
    global conn
    conn.baremetal.set_node_provision_state(node, state)


@app.task(bind=True, name="osism.tasks.openstack.baremetal_create_allocations")
def baremetal_create_allocations(self, nodes):
    global conn

    for node in nodes:
        try:
            allocation_a = conn.baremetal.get_allocation(allocation=node)
        except openstack.exceptions.ResourceNotFound:
            allocation_a = None

        if not allocation_a:
            # Get Ironic parameters from the conductor
            task = conductor.get_ironic_parameters.delay()
            task.wait(timeout=None, interval=0.5)
            ironic_parameters = task.get()

            allocation_a = conn.baremetal.create_allocation(
                name=node,
                candidate_nodes=[node],
                resource_class=ironic_parameters["resource_class"],
            )
            conn.baremetal.wait_for_allocation(allocation=node, timeout=30)


@app.task(bind=True, name="osism.tasks.openstack.baremetal_create_nodes")
def baremetal_create_nodes(self, nodes, ironic_parameters):
    global conn

    for node in nodes:
        # TODO: Filter on mgmt_only
        address_a = utils.nb.ipam.ip_addresses.get(device=node, interface="Ethernet0")

        node_parameters = copy.deepcopy(ironic_parameters)

        if node_parameters["driver"] == "redfish":
            remote_board_address = str(ipaddress.ip_interface(address_a["address"]).ip)
            t = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(
                node_parameters["driver_info"]["redfish_address"]
            )
            node_parameters["driver_info"]["redfish_address"] = t.render(
                remote_board_address=remote_board_address
            )

        elif node_parameters["driver"] == "ipmi":
            remote_board_address = str(ipaddress.ip_interface(address_a["address"]).ip)
            t = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(
                node_parameters["driver_info"]["ipmi_address"]
            )
            node_parameters["driver_info"]["ipmi_address"] = t.render(
                remote_board_address=remote_board_address
            )

        try:
            device_a = utils.nb.dcim.devices.get(name=node)
            tags = [str(tag) for tag in device_a.tags]

            # NOTE: Internally used nodes are identified by their unique name via the resource class.
            #       The actual resource class is explicitly overwritten.
            if "Managed by Ironic" in tags and "Managed by OSISM" in tags:
                node_parameters["resource_class"] = f"osism-{node}"
                baremetal_create_internal_flavor(node)

            conn.baremetal.create_node(
                name=node, provision_state="manageable", **node_parameters
            )
            conn.baremetal.wait_for_nodes_provision_state([node], "manageable")

            if "Managed by Ironic" in tags and "Managed by OSISM" not in tags:
                conn.baremetal.set_node_traits(node, ["CUSTOM_GENERAL_USE"])
            elif "Managed by Ironic" in tags and "Managed by OSISM" in tags:
                conn.baremetal.set_node_traits(node, ["CUSTOM_OSISM_USE"])

            conn.baremetal.set_node_provision_state(node, "inspect")

            # TODO: Check if the system has been registered correctly
            device_a.custom_fields = {
                "ironic_state": "registered",
            }
            device_a.save()

        except openstack.exceptions.ResourceFailure:
            # TODO: Do something useful here
            pass
        except openstack.exceptions.ConflictException:
            # The node already exists and has a wronge state in the Netbox
            device_a = utils.nb.dcim.devices.get(name=node)
            device_a.custom_fields = {
                "ironic_state": "registered",
            }
            device_a.save()


@app.task(bind=True, name="osism.tasks.openstack.baremetal_check_allocations")
def baremetal_check_allocations(self):
    lock = Redlock(
        key="lock_osism_tasks_openstack_baremetal_check_allocations",
        masters={redis},
        auto_release_time=60,
    )

    if lock.acquire(timeout=20):
        netbox.get_devices_that_should_have_an_allocation_in_ironic.apply_async(
            (), link=baremetal_create_allocations.s()
        )
        lock.release()


@app.task(bind=True, name="osism.tasks.openstack.baremetal_create_internal_flavor")
def baremetal_create_internal_flavor(self, node):
    flavor_a = conn.compute.create_flavor(
        name=f"osism-{node}", ram=1, vcpus=1, disk=1, is_public=False
    )
    specs = {
        f"resources:CUSTOM_RESOURCE_CLASS_OSISM_{node.upper()}": 1,
        "resources:VCPU": 0,
        "resources:MEMORY_MB": 0,
        "resources:DISK_GB": 0,
        "trait:CUSTOM_OSISM_USE": "required",
    }
    conn.compute.create_flavor_extra_specs(flavor_a, specs)


@app.task(bind=True, name="osism.tasks.openstack.baremetal_delete_internal_flavor")
def baremetal_delete_internal_flavor(self, node):
    flavor = conn.compute.get_flavor(f"osism-{node}")
    conn.compute.delete_flavor(flavor)



# SPDX-License-Identifier: Apache-2.0

import os
import re
import subprocess
import time

from celery.signals import worker_process_init
from loguru import logger
from redis import Redis
from pottery import Redlock

from osism import settings

redis = None


class Config:
    broker_connection_retry_on_startup = True
    enable_utc = True
    enable_ironic = os.environ.get("ENABLE_IRONIC", "True")
    broker_url = "redis://redis"
    result_backend = "redis://redis"
    task_create_missing_queues = True
    task_default_queue = "default"
    task_track_started = (True,)
    task_routes = {
        "osism.tasks.ansible.*": {"queue": "osism-ansible"},
        "osism.tasks.ceph.*": {"queue": "ceph-ansible"},
        "osism.tasks.conductor.*": {"queue": "conductor"},
        "osism.tasks.kolla.*": {"queue": "kolla-ansible"},
        "osism.tasks.kubernetes.*": {"queue": "kubernetes"},
        "osism.tasks.netbox.*": {"queue": "netbox"},
        "osism.tasks.openstack.*": {"queue": "openstack"},
        "osism.tasks.reconciler.*": {"queue": "reconciler"},
    }


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global redis

    redis = Redis(
        host=settings.REDIS_HOST,
        port=settings.REDIS_PORT,
        db=settings.REDIS_DB,
        socket_keepalive=True,
    )
    redis.ping()


def run_ansible_in_environment(
    request_id,
    worker,
    environment,
    role,
    arguments,
    publish=True,
    locking=False,
    auto_release_time=3600,
):
    result = ""

    if type(arguments) == list:
        joined_arguments = " ".join(arguments)
    else:
        joined_arguments = arguments

    env = os.environ.copy()

    # Bring back colored Ansible output, thanks to
    # https://www.jeffgeerling.com/blog/2020/getting-colorized-output-molecule-and-ansible-on-github-actions-ci
    env["ANSIBLE_FORCE_COLOR"] = "1"
    env["PY_COLORS"] = "1"

    # handle sub environments
    if "." in environment:
        sub_name = environment.split(".")[1]
        env["SUB"] = environment
        environment = environment.split(".")[0]
        logger.info(
            f"worker = {worker}, environment = {environment}, sub = {sub_name}, role = {role}"
        )
    else:
        logger.info(f"worker = {worker}, environment = {environment}, role = {role}")

    env["ENVIRONMENT"] = environment

    # NOTE: This is a first step to make Ansible Vault usable via OSISM workers.
    #       It's not ready in that form yet.
    ansible_vault_password = redis.get("ansible_vault_password")
    if ansible_vault_password:
        env["VAULT"] = "/ansible-vault.py"

    # NOTE: Consider arguments in the future
    if locking:
        lock = Redlock(
            key=f"lock-ansible-{environment}-{role}",
            masters={redis},
            auto_release_time=auto_release_time,
        )

    # NOTE: use python interface in the future, something with ansible-runner and the fact cache is
    #       not working out of the box

    # execute roles from kolla-ansible
    if worker == "kolla-ansible":
        if locking:
            lock.acquire()

        if role in ["mariadb-backup", "mariadb_backup"]:
            action = "backup"
            role = "mariadb"
            # Hacky workaround. The handling of kolla_action will be revised in the future.
            joined_arguments = re.sub(
                r"-e kolla_action=(bootstrap|config|deploy|precheck|pull|reconfigure|refresh-containers|stop|upgrade)",
                r"-e kolla_action=backup",
                joined_arguments,
            )
        else:
            action = "deploy"

        command = f"/run.sh {action} {role} {joined_arguments}"
        logger.info(f"RUN {command}")
        p = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            env=env,
        )

    # execute roles from kubernetes
    elif worker == "kubernetes":
        if locking:
            lock.acquire()

        command = f"/run.sh {role} {joined_arguments}"
        logger.info(f"RUN {command}")
        p = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            env=env,
        )

    # execute roles from ceph-ansible
    elif worker == "ceph-ansible":
        if locking:
            lock.acquire()

        command = f"/run.sh {role} {joined_arguments}"
        logger.info(f"RUN {command}")
        p = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            env=env,
        )

    # execute local netbox playbooks
    elif worker == "osism-ansible" and environment == "netbox-local":
        if locking:
            lock.acquire()

        command = f"/run-{environment}.sh {role} {joined_arguments}"
        logger.info(f"RUN {command}")
        p = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
        )

    # execute all other roles
    else:
        if locking:
            lock.acquire()

        command = f"/run-{environment}.sh {role} {joined_arguments}"
        logger.info(f"RUN {command}")
        p = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
            env=env,
        )

    while p.poll() is None:
        line = p.stdout.readline().decode("utf-8")
        if publish:
            redis.xadd(request_id, {"type": "stdout", "content": line})
        result += line

    rc = p.wait(timeout=60)

    if publish:
        redis.xadd(request_id, {"type": "rc", "content": rc})
        redis.xadd(request_id, {"type": "action", "content": "quit"})

    if locking:
        lock.release()

    return result


def handle_task(t, wait, format, timeout):
    global redis

    if not redis:
        redis = Redis(
            host=settings.REDIS_HOST,
            port=settings.REDIS_PORT,
            db=settings.REDIS_DB,
            socket_keepalive=True,
        )
        redis.ping()

    rc = 0
    if wait:
        stoptime = time.time() + timeout
        last_id = 0
        while time.time() < stoptime:
            data = redis.xread(
                {str(t.task_id): last_id}, count=1, block=(timeout * 1000)
            )
            if data:
                stoptime = time.time() + timeout
                messages = data[0]
                for message_id, message in messages[1]:
                    last_id = message_id.decode()
                    message_type = message[b"type"].decode()
                    message_content = message[b"content"].decode()

                    logger.debug(f"Processing message {last_id} of type {message_type}")
                    redis.xdel(str(t.task_id), last_id)

                    if message_type == "stdout":
                        print(message_content, end="", flush=True)
                        if "PLAY RECAP" in message_content:
                            logger.info(
                                "Play has been completed. There may now be a delay until "
                                "all logs have been written."
                            )
                            logger.info("Please wait and do not abort execution.")
                    elif message_type == "rc":
                        rc = int(message_content)
                    elif message_type == "action" and message_content == "quit":
                        redis.close()
                        return rc
        else:
            logger.info(
                f"There has been no output from the task {t.task_id} for {timeout} second(s)."
            )
            logger.info(
                f"The task timeout of {timeout} second(s) can be adjusted using the --timeout parameter."
            )
            logger.info(
                f"Task {t.task_id} is still running in background. Check ARA for further logs. "
            )
            logger.info(
                "Use this command to continue waiting for this task: "
                f"osism wait --output --live --delay 2 {t.task_id}"
            )
            return 1

    else:
        if format == "log":
            logger.info(
                f"Task {t.task_id} is running in background. No more output. Check ARA for logs."
            )
        elif format == "script":
            print(f"{t.task_id}")

        return rc



# SPDX-License-Identifier: Apache-2.0

LOADBALANCER_PLAYBOOKS = [
    "loadbalancer-aodh",
    "loadbalancer-barbican",
    "loadbalancer-blazar",
    "loadbalancer-ceph-rgw",
    "loadbalancer-cinder",
    "loadbalancer-designate",
    "loadbalancer-glance",
    "loadbalancer-gnocchi",
    "loadbalancer-grafana",
    "loadbalancer-heat",
    "loadbalancer-horizon",
    "loadbalancer-ironic",
    "loadbalancer-keystone",
    "loadbalancer-magnum",
    "loadbalancer-manila",
    "loadbalancer-mariadb",
    "loadbalancer-memcached",
    "loadbalancer-neutron",
    "loadbalancer-nova",
    "loadbalancer-octavia",
    "loadbalancer-opensearch",
    "loadbalancer-placement",
    "loadbalancer-prometheus",
    "loadbalancer-rabbitmq",
    "loadbalancer-skydive",
]

VALIDATE_PLAYBOOKS = {
    "barbican-config": {
        "runtime": "kolla-ansible",
        "playbook": "barbican",
    },
    "blazar-config": {
        "runtime": "kolla-ansible",
        "playbook": "blazar",
    },
    "designate-config": {
        "runtime": "kolla-ansible",
        "playbook": "designate",
    },
    "keystone-config": {
        "runtime": "kolla-ansible",
        "playbook": "keystone",
    },
    "glance-config": {
        "runtime": "kolla-ansible",
        "playbook": "glance",
    },
    "heat-config": {
        "runtime": "kolla-ansible",
        "playbook": "heat",
    },
    "octavia-config": {
        "runtime": "kolla-ansible",
        "playbook": "octavia",
    },
    "nova-config": {
        "runtime": "kolla-ansible",
        "playbook": "nova",
    },
    "neutron-config": {
        "runtime": "kolla-ansible",
        "playbook": "neutron",
    },
    "placement-config": {
        "runtime": "kolla-ansible",
        "playbook": "placement",
    },
    "aodh-config": {
        "runtime": "kolla-ansible",
        "playbook": "aodh",
    },
    "ceilometer-config": {
        "runtime": "kolla-ansible",
        "playbook": "ceilometer",
    },
    "cloudkitty-config": {
        "runtime": "kolla-ansible",
        "playbook": "cloudkitty",
    },
    "ironic-config": {
        "runtime": "kolla-ansible",
        "playbook": "ironic",
    },
    "manila-config": {
        "runtime": "kolla-ansible",
        "playbook": "manila",
    },
    "senlin-config": {
        "runtime": "kolla-ansible",
        "playbook": "senlin",
    },
    # NOTE: The command should be "osism validate ceph-config". However,
    # the corresponding playbook is called ceph-validate because ceph-config
    # deploys the Ceph configuration itself. So this is rewritten from
    # ceph-config to ceph-validate.
    "ceph-config": {
        "runtime": "ceph-ansible",
        "playbook": "validate",
    },
    # NOTE: The playbooks for validating the Ceph deployment are currently
    # in osism/ansible-playbooks. Therefore, they are not executed in
    # ceph-ansible but in osism-ansible.
    "ceph-connectivity": {"environment": "ceph", "runtime": "osism-ansible"},
    "ceph-mgrs": {"environment": "ceph", "runtime": "osism-ansible"},
    "ceph-mons": {"environment": "ceph", "runtime": "osism-ansible"},
    "ceph-osds": {"environment": "ceph", "runtime": "osism-ansible"},
    "container-status": {"environment": "generic", "runtime": "osism-ansible"},
    "kernel-version": {"environment": "generic", "runtime": "osism-ansible"},
    "kolla-connectivity": {"environment": "kolla", "runtime": "osism-ansible"},
    "mysql-open-files-limit": {"environment": "generic", "runtime": "osism-ansible"},
    "ntp": {"environment": "generic", "runtime": "osism-ansible"},
    "system-encoding": {"environment": "generic", "runtime": "osism-ansible"},
    "ulimits": {"environment": "generic", "runtime": "osism-ansible"},
    "mariadb-backup": {"environment": "kolla", "runtime": "kolla-ansible"},
    "mariadb-recovery": {"environment": "kolla", "runtime": "kolla-ansible"},
}

MAP_ROLE2ROLE = {
    "nutshell": [
        "dotfiles",
        "homer",
        "netdata",
        "openstackclient",
        "phpmyadmin",
        [
            "common",
            [
                [
                    "loadbalancer",
                    [
                        "opensearch",
                        [
                            "mariadb-ng",
                            [
                                "horizon",
                                [
                                    "keystone",
                                    [
                                        ["neutron", ["octavia"]],
                                        "barbican",
                                        "designate",
                                        "ironic",
                                        "placement",
                                        "magnum",
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
                ["openvswitch", ["ovn"]],
                "memcached",
                "redis",
                "rabbitmq-ng",
            ],
        ],
        ["kubernetes", ["kubeconfig", ["copy-kubeconfig"]]],
        [
            "ceph",
            [
                [
                    "ceph-pools",
                    [
                        [
                            "copy-ceph-keys",
                            [
                                "cephclient",
                                "ceph-bootstrap-dashboard",
                                [
                                    "wait-for-keystone",
                                    ["kolla-ceph-rgw", "glance", "cinder", "nova"],
                                ],
                                ["prometheus", ["grafana"]],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
    "collection-infrastructure": [
        "openstackclient",
        "phpmyadmin",
        [
            "common",
            [
                ["loadbalancer", ["opensearch", "mariadb-ng"]],
                ["openvswitch", ["ovn"]],
                "memcached",
                "redis",
                "rabbitmq-ng",
            ],
        ],
    ],
    "collection-kubernetes": [
        ["kubernetes", ["kubeconfig", ["copy-kubeconfig"]]],
    ],
    "collection-openstack": [
        "horizon",
        [
            "keystone",
            [
                "glance",
                "cinder",
                "neutron",
                "barbican",
                "designate",
                "octavia",
                "ironic",
                "kolla-ceph-rgw",
                "magnum",
                ["placement", ["nova"]],
            ],
        ],
    ],
    "collection-ceph": [
        [
            "ceph",
            [
                [
                    "ceph-pools",
                    [["copy-ceph-keys", ["cephclient", "ceph-bootstrap-dashboard"]]],
                ],
            ],
        ],
    ],
    "collection-monitoring": [
        ["prometheus", ["grafana"]],
        "netdata",
    ],
    "collection-bootstrap": [
        [
            "gather-facts",
            [
                [
                    "hostname",
                    [
                        [
                            "hosts",
                            [
                                [
                                    "proxy",
                                    [
                                        [
                                            "resolvconf",
                                            [
                                                [
                                                    "repository",
                                                    [
                                                        "rsyslog",
                                                        "journald",
                                                        "systohc",
                                                        "configfs",
                                                        "packages",
                                                        "sysctl",
                                                        "limits",
                                                        "services",
                                                        "motd",
                                                        "rng",
                                                        "smartd",
                                                        "cleanup",
                                                        "timezone",
                                                        "docker",
                                                        "docker-compose",
                                                        "chrony",
                                                        "lldpd",
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
}



# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
import pprint
import subprocess

from celery import Celery
from cliff.command import Command
import docker
import json
from loguru import logger
from tabulate import tabulate

from osism.tasks import Config
from osism.utils import redis


class VersionsManager(Command):
    def get_parser(self, prog_name):
        parser = super(VersionsManager, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        client = docker.from_env()

        data = []

        for cname in ["osism-ansible", "ceph-ansible", "kolla-ansible"]:
            try:
                container = client.containers.get(cname)
                version = container.labels["org.opencontainers.image.version"]

                if cname == "ceph-ansible":
                    mrelease = container.labels["de.osism.release.ceph"]
                elif cname == "kolla-ansible":
                    mrelease = container.labels["de.osism.release.openstack"]
                else:
                    mrelease = ""

                data.append([cname, version, mrelease])
            except docker.errors.NotFound:
                pass

        result = tabulate(
            data, headers=["Module", "OSISM version", "Module release"], tablefmt="psql"
        )
        print(result)


class Tasks(Command):
    def get_parser(self, prog_name):
        parser = super(Tasks, self).get_parser(prog_name)
        parser.add_argument(
            "--status", default="all", help="Status of the tasks to list"
        )
        return parser

    def take_action(self, parsed_args):
        status = parsed_args.status

        app = Celery("task")
        app.config_from_object(Config)

        i = app.control.inspect()

        table = []

        task_status = "ACTIVE"
        for worker, tasks in i.active().items():
            for task in tasks:
                time_start = datetime.fromtimestamp(task["time_start"])
                table.append(
                    [
                        worker,
                        task["id"],
                        task["name"],
                        task_status,
                        time_start,
                        task["args"],
                    ]
                )

        task_status = "SCHEDULED"
        for worker, tasks in i.scheduled().items():
            for task in tasks:
                time_start = datetime.fromtimestamp(task["time_start"])
                table.append(
                    [
                        worker,
                        task["id"],
                        task["name"],
                        task_status,
                        time_start,
                        task["args"],
                    ]
                )

        print(
            tabulate(
                table,
                headers=["Worker", "ID", "Name", "Status", "Start time", "Arguments"],
                tablefmt="psql",
            )
        )


class Hostvars(Command):
    def get_parser(self, prog_name):
        parser = super(Hostvars, self).get_parser(prog_name)
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Hostname (as the host is known in Ansible inventory)",
        )
        parser.add_argument(
            "variable",
            nargs="?",
            type=str,
            help="Name of a variable to show",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        variable = parsed_args.variable

        try:
            result = subprocess.check_output(
                f"ansible-inventory -i /ansible/inventory/hosts.yml --host {host}",
                shell=True,
                stderr=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError:
            logger.error(f"Host {host} not found in inventory.")
            return

        data = json.loads(result)
        table = []

        if variable:
            if variable in data:
                row = pprint.pformat(data[variable], indent=2, width=60, compact=True)
                table.append([host, variable, row])
            else:
                logger.error(f"Variable {variable} not found in inventory for {host}.")
        else:
            for variable in data:
                row = pprint.pformat(data[variable], indent=2, width=60, compact=True)
                table.append([host, variable, row])

        if table:
            print(
                tabulate(table, headers=["Host", "Variable", "Value"], tablefmt="grid")
            )

        return


class Facts(Command):
    def get_parser(self, prog_name):
        parser = super(Facts, self).get_parser(prog_name)
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Hostname (as the host is known in Ansible inventory)",
        )
        parser.add_argument(
            "fact",
            nargs="?",
            type=str,
            help="Name of a fact to show",
        )
        parser.add_argument(
            "--no-cache",
            default=False,
            help="Do not use facts from the cache",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        fact = parsed_args.fact
        cache = not parsed_args.no_cache

        data = redis.get(f"ansible_facts{host}")
        if data:
            data = json.loads(data)
            table = []

            if fact:
                if fact in data:
                    row = pprint.pformat(data[fact], indent=2, width=60, compact=True)
                    table.append([host, fact, row])
                else:
                    logger.error(f"Fact {fact} not found in cache for {host}.")
            else:
                for fact in data:
                    row = pprint.pformat(data[fact], indent=2, width=60, compact=True)
                    if fact in [
                        "ansible_ssh_host_key_dsa_public",
                        "ansible_ssh_host_key_ecdsa_public",
                        "ansible_ssh_host_key_ed25519_public",
                        "ansible_ssh_host_key_rsa_public",
                    ]:
                        row = f"{row[0:40]}..."
                    table.append([host, fact, row])

            if table:
                print(
                    tabulate(table, headers=["Host", "Fact", "Value"], tablefmt="grid")
                )
        else:
            logger.error(f"No facts found in cache for {host}.")

        return


class Hosts(Command):
    def get_parser(self, prog_name):
        parser = super(Hosts, self).get_parser(prog_name)
        parser.add_argument(
            "-l",
            "--limit",
            type=str,
            help="Limit selected hosts to an additional pattern",
        )
        return parser

    def take_action(self, parsed_args):
        try:
            command = ["ansible-inventory -i /ansible/inventory/hosts.yml --list"]
            if parsed_args.limit:
                command.append(f"--limit {parsed_args.limit}")

            result = subprocess.check_output(
                " ".join(command),
                shell=True,
                stderr=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError:
            logger.error("Error loading inventory.")
            return

        data = json.loads(result)
        table = []

        for host in data["_meta"]["hostvars"]:
            table.append([host])

        if table:
            print(tabulate(table, headers=["Host"], tablefmt="psql"))

        return


class States(Command):
    def get_parser(self, prog_name):
        parser = super(States, self).get_parser(prog_name)
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Hostname (as the host is known in Ansible inventory)",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]

        data = redis.get(f"ansible_facts{host}")
        if data:
            data = json.loads(data)
            table = []

            if "ansible_local" in data and "osism" in data["ansible_local"]:
                for local_fact in data["ansible_local"]["osism"]:
                    if local_fact != "bootstrap":
                        local_data = data["ansible_local"]["osism"][local_fact]
                        table.append(
                            [local_fact, local_data["state"], local_data["timestamp"]]
                        )

                if table:
                    print(
                        tabulate(
                            table,
                            headers=["Role", "State", "Timestamp"],
                            tablefmt="psql",
                        )
                    )

        return



# SPDX-License-Identifier: Apache-2.0

# NOTE: This is a first step to make Ansible Vault usable via OSISM workers.
#       It's not ready in that form yet.

import os
import subprocess

from cliff.command import Command
from cryptography.fernet import Fernet
from prompt_toolkit import prompt

from osism.utils import redis


class SetPassword(Command):
    keyfile = "/share/ansible_vault_password.key"

    def get_parser(self, prog_name):
        parser = super(SetPassword, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        if os.path.isfile(self.keyfile):
            with open(self.keyfile, "r") as fp:
                key = fp.read()
        else:
            key = Fernet.generate_key()
            with open(self.keyfile, "w+") as fp:
                fp.write(key.decode("utf-8"))

        f = Fernet(key)

        ansible_vault_password = prompt("Ansible Vault password: ", is_password=True)
        redis.set(
            "ansible_vault_password", f.encrypt(ansible_vault_password.encode("utf-8"))
        )


class UnsetPassword(Command):
    def get_parser(self, prog_name):
        parser = super(UnsetPassword, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        redis.delete("ansible_vault_password")


class View(Command):
    def get_parser(self, prog_name):
        parser = super(View, self).get_parser(prog_name)
        parser.add_argument(
            "path", nargs="?", type=str, help="Path to the secret.yml file"
        )
        return parser

    def take_action(self, parsed_args):
        path = parsed_args.path
        if not os.path.isabs(path):
            path = os.path.join("/opt/configuration", path)
        subprocess.call(f"/usr/local/bin/ansible-vault view {path}", shell=True)



# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path
from typing import Dict, Any

from loguru import logger
import yaml

MAP_ROLE2ENVIRONMENT: Dict[str, Any] = {}
MAP_ROLE2RUNTIME: Dict[str, Any] = {}

for path in Path("/interface/playbooks").glob("*.yml"):
    try:
        with open(path) as fp:
            data = yaml.load(fp, Loader=yaml.SafeLoader)

        MAP_ROLE2ENVIRONMENT = MAP_ROLE2ENVIRONMENT | data
        MAP_ROLE2RUNTIME[os.path.basename(path)[:-4]] = data.keys()

    # Ignore YAML errors here so that we can prevent the command
    # from being non-functional if a runtime environment provides
    # an invalid interface file.
    except yaml.YAMLError as e:
        logger.warning(e)



# SPDX-License-Identifier: Apache-2.0

import os
from re import findall
import subprocess
from urllib.parse import urljoin

from cliff.command import Command
import docker
from jinja2 import Template
from loguru import logger
import requests

from osism.data import TEMPLATE_IMAGE_CLUSTERAPI, TEMPLATE_IMAGE_OCTAVIA

SUPPORTED_CLUSTERAPI_K8S_IMAGES = ["1.28", "1.29", "1.30", "1.31"]


class ImageClusterapi(Command):
    def get_parser(self, prog_name):
        parser = super(ImageClusterapi, self).get_parser(prog_name)

        parser.add_argument(
            "--base-url",
            type=str,
            help="Base URL",
            default="https://swift.services.a.regiocloud.tech/swift/v1/AUTH_b182637428444b9aa302bb8d5a5a418c/openstack-k8s-capi-images/",
        )
        parser.add_argument(
            "--cloud",
            type=str,
            help="Cloud name in clouds.yaml (will be overruled by OS_AUTH_URL envvar)",
            default="admin",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Do not perform any changes (--dry-run passed to openstack-image-manager)",
        )
        parser.add_argument(
            "--tag",
            type=str,
            help="Name of the tag used to identify managed images (use openstack-image-manager's default if unset)",
            default=None,
        )
        parser.add_argument(
            "--filter",
            type=str,
            help="Filter the version to be managed (e.g. 1.28)",
            default=None,
        )
        return parser

    def take_action(self, parsed_args):
        base_url = parsed_args.base_url
        cloud = parsed_args.cloud
        filter = parsed_args.filter
        tag = parsed_args.tag

        if filter:
            supported_cluterapi_k8s_images = [filter]
        else:
            supported_cluterapi_k8s_images = SUPPORTED_CLUSTERAPI_K8S_IMAGES

        os.makedirs("/tmp/clusterapi", exist_ok=True)
        for kubernetes_release in supported_cluterapi_k8s_images:
            url = urljoin(base_url, f"last-{kubernetes_release}")

            response = requests.get(url)
            splitted = response.text.strip().split(" ")

            logger.info(f"date: {splitted[0]}")
            logger.info(f"image: {splitted[1]}")

            r = findall(r".*ubuntu-2204-kube-v(.*\..*\..*).qcow2", splitted[1])
            logger.info(f"version: {r[0].strip()}")

            url = urljoin(base_url, splitted[1])
            logger.info(f"url: {url}")

            logger.info(f"checksum_url: {url}.CHECKSUM")
            response_checksum = requests.get(f"{url}.CHECKSUM")
            splitted_checksum = response_checksum.text.strip().split(" ")
            logger.info(f"checksum: {splitted_checksum[0]}")

            template = Template(TEMPLATE_IMAGE_CLUSTERAPI)
            result = template.render(
                image_url=url,
                image_checksum=f"sha256:{splitted_checksum[0]}",
                image_version=r[0].strip(),
                image_builddate=splitted[0],
            )
            with open(f"/tmp/clusterapi/k8s-{kubernetes_release}.yml", "w+") as fp:
                fp.write(result)

        args = [
            "openstack-image-manager",
            "--images=/tmp/clusterapi",
            "--cloud",
            cloud,
            "--filter",
            "ubuntu-capi-image",
        ]
        if tag is not None:
            args.extend(["--tag", tag])
        if parsed_args.dry_run:
            args.append("--dry-run")
        subprocess.call(args)


class ImageOctavia(Command):
    def get_parser(self, prog_name):
        parser = super(ImageOctavia, self).get_parser(prog_name)

        parser.add_argument(
            "--cloud", type=str, help="Cloud name in clouds.yaml", default="openstack"
        )
        parser.add_argument(
            "--base-url",
            type=str,
            help="Base URL",
            default="https://swift.services.a.regiocloud.tech/swift/v1/AUTH_b182637428444b9aa302bb8d5a5a418c/openstack-octavia-amphora-image/",
        )
        return parser

    def take_action(self, parsed_args):
        cloud = parsed_args.cloud
        base_url = parsed_args.base_url

        client = docker.from_env()
        container = client.containers.get("kolla-ansible")
        openstack_release = container.labels["de.osism.release.openstack"]
        url = urljoin(base_url, f"last-{openstack_release}")

        response = requests.get(url)
        splitted = response.text.strip().split(" ")

        logger.info(f"date: {splitted[0]}")
        logger.info(f"image: {splitted[1]}")

        url = urljoin(base_url, splitted[1])
        logger.info(f"url: {url}")

        logger.info(f"checksum_url: {url}.CHECKSUM")
        response_checksum = requests.get(f"{url}.CHECKSUM")
        splitted_checksum = response_checksum.text.strip().split(" ")
        logger.info(f"checksum: {splitted_checksum[0]}")

        template = Template(TEMPLATE_IMAGE_OCTAVIA)
        result = template.render(
            image_url=url,
            image_checksum=f"sha256:{splitted_checksum[0]}",
            image_version=splitted[0],
            image_builddate=splitted[0],
        )

        os.makedirs("/tmp/octavia", exist_ok=True)
        with open("/tmp/octavia/octavia.yml", "w+") as fp:
            fp.write(result)

        subprocess.call(
            "/usr/local/bin/openstack-image-manager --images=/tmp/octavia --cloud octavia --deactivate",
            shell=True,
        )


class Images(Command):
    def get_parser(self, prog_name):
        parser = super(Images, self).get_parser(prog_name)

        # NOTE: This is a workaround. argparse.REMAINDER does not work well to pass all
        # arguments to openstack-image-manager. This is improved by switching from cliff
        # to typer. Then openstack-image-manager can simply be included directly at this
        # point.

        parser.add_argument(
            "--dry-run",
            default=False,
            help="Do not perform any changes",
            action="store_true",
        )
        parser.add_argument(
            "--hide",
            default=True,
            help="Hide images that should be deleted",
            action="store_true",
        )
        parser.add_argument(
            "--latest",
            default=False,
            help="Only import the latest version for images of type multi",
            action="store_true",
        )
        parser.add_argument(
            "--cloud", type=str, help="Cloud name in clouds.yaml", default="openstack"
        )
        parser.add_argument(
            "--filter",
            type=str,
            help="Filter images with a regex on their name",
            default=None,
        )
        parser.add_argument(
            "--images",
            type=str,
            help="Path to the directory containing all image files or path to single image file",
            default=None,
        )

        return parser

    def take_action(self, parsed_args):
        cloud = parsed_args.cloud
        dry_run = parsed_args.dry_run
        filter = parsed_args.filter
        hide = parsed_args.hide
        latest = parsed_args.latest
        images = parsed_args.images

        arguments = []
        if cloud:
            arguments.append(f"--cloud '{cloud}'")
        if filter:
            arguments.append(f"--filter '{filter}'")
        if dry_run:
            arguments.append("--dry-run")
        if latest:
            arguments.append("--latest")
        if hide:
            arguments.append("--hide")
        if images:
            arguments.append(f"--images '{images}'")

        joined_arguments = " ".join(arguments)
        subprocess.call(
            f"/usr/local/bin/openstack-image-manager --images=/etc/images {joined_arguments}",
            shell=True,
        )


class Flavors(Command):
    def get_parser(self, prog_name):
        parser = super(Flavors, self).get_parser(prog_name)

        # NOTE: This is a workaround. argparse.REMAINDER does not work well to pass all
        # arguments to openstack-image-manager. This is improved by switching from cliff
        # to typer. Then openstack-flavor-manager can simply be included directly at this
        # point.

        parser.add_argument(
            "--cloud", type=str, help="Cloud name in clouds.yaml", default="admin"
        )
        parser.add_argument(
            "--name",
            type=str,
            help="Name of flavor definitions",
            default="scs",
            choices=["scs", "osism", "local", "url"],
        )
        parser.add_argument(
            "--url",
            type=str,
            help="Overwrite the default URL where the flavor definitions are available",
            default=None,
        )
        parser.add_argument(
            "--recommended",
            default=False,
            help="Also create recommended flavors",
            action="store_true",
        )

        return parser

    def take_action(self, parsed_args):
        cloud = parsed_args.cloud
        name = parsed_args.name
        recommended = parsed_args.recommended
        url = parsed_args.url

        arguments = [f"--name '{name}'"]
        if cloud:
            arguments.append(f"--cloud '{cloud}'")

        if recommended:
            arguments.append("--recommended")

        if url:
            arguments.append(f"--url '{url}'")

        joined_arguments = " ".join(arguments)
        subprocess.call(
            f"/usr/local/bin/openstack-flavor-manager {joined_arguments}",
            shell=True,
        )



# SPDX-License-Identifier: Apache-2.0

from celery import Celery

from osism.tasks import Config, run_ansible_in_environment

app = Celery("kolla")
app.config_from_object(Config)


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    pass


@app.task(bind=True, name="osism.tasks.kolla.run")
def run(self, environment, playbook, arguments, publish=True, auto_release_time=3600):
    return run_ansible_in_environment(
        self.request.id,
        "kolla-ansible",
        environment,
        playbook,
        arguments,
        publish,
        False,
        auto_release_time,
    )



# SPDX-License-Identifier: Apache-2.0

from loguru import logger
from pottery import Redlock
import git
import jinja2

from osism import utils


def for_device(name, parameters={}, mode="deploy"):
    device = utils.nb.dcim.devices.get(name=name)

    if (
        "device_type" not in device.custom_fields
        or device.custom_fields["device_type"] != "switch"
    ):
        return

    if "Managed by OSISM" not in [str(x) for x in device.tags]:
        return

    if "deployment_enabled" in device.custom_fields and not bool(
        device.custom_fields["deployment_enabled"]
    ):
        return

    if "deployment_type" not in device.custom_fields:
        return

    # Allow only one change per time
    lock = Redlock(
        key=f"lock_deploy_{name}", masters={utils.redis}, auto_release_time=120
    )
    lock.acquire()

    repo = git.Repo.init(path="/state")

    first = False

    if device.name in repo.tags:
        last_commit = repo.commit(device.name)
        current_commit = repo.head.commit
    else:
        first = True
        last_commit = repo.head.commit
        current_commit = repo.head.commit

    if device.name in repo.tags and mode == "deploy":
        repo.delete_tag(name)

    if not first and last_commit == current_commit and mode == "deploy":
        logger.info(f"No deployment for device {device.name} required")
    else:
        if not first:
            try:
                last_configuration = repo.git.show(
                    f"{last_commit.hexsha}:{device.name}.cfg.j2"
                )
            except git.exc.GitCommandError:
                last_configuration = None
        else:
            last_configuration = None

        try:
            current_configuration = repo.git.show(
                f"{current_commit.hexsha}:{device.name}.cfg.j2"
            )
        except git.exc.GitCommandError:
            current_configuration = None

        if not current_configuration:
            logger.error(f"There is no prepared configuration for device {device.name}")
        else:
            t = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(
                current_configuration
            )
            rendered_current_configuration = t.render(**parameters)

            logger.info(
                f"{mode} configuration for device {device.name} with plugin {device.custom_fields['deployment_type']}"
            )

            deployment_type = device.custom_fields["deployment_type"]
            logger.error(
                f"Deployment type {deployment_type} for device {device.name} not supported"
            )

    if mode == "deploy":
        repo.create_tag(device.name, current_commit)

    lock.release()



# SPDX-License-Identifier: Apache-2.0

import argparse
from glob import glob
from os.path import basename

from cliff.command import Command
from loguru import logger
from redis import Redis
from tabulate import tabulate

from osism import settings
from osism.tasks import conductor, netbox, reconciler, ansible, openstack, handle_task


redis = Redis(host=settings.REDIS_HOST, port=settings.REDIS_PORT, db=settings.REDIS_DB)
redis.ping()


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Other arguments for Ansible"
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the role has been applied",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        pass


class Ironic(Command):
    def get_parser(self, prog_name):
        parser = super(Ironic, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        # Get Ironic parameters from the conductor
        task = conductor.get_ironic_parameters.delay()
        task.wait(timeout=None, interval=0.5)
        ironic_parameters = task.get()

        # Add all unregistered systems from the Netbox in Ironic
        netbox.get_devices_not_yet_registered_in_ironic.apply_async(
            (), link=openstack.baremetal_create_nodes.s(ironic_parameters)
        )

        # Synchronize the current status in Ironic with the Netbox
        # openstack.baremetal_node_list.apply_async((), link=netbox.synchronize_device_state.s())

        # Remove systems from Ironic that are no longer present in the Netbox


class Sync(Command):
    def get_parser(self, prog_name):
        parser = super(Sync, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the sync has been completed",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait

        task = reconciler.sync_inventory_with_netbox.delay()
        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Init(Command):
    def get_parser(self, prog_name):
        parser = super(Init, self).get_parser(prog_name)
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Other arguments for Ansible"
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the role has been applied",
            action="store_true",
        )
        parser.add_argument(
            "--format",
            default="log",
            help="Output type",
            const="log",
            nargs="?",
            choices=["script", "log"],
        ),
        return parser

    def take_action(self, parsed_args):
        arguments = parsed_args.arguments
        format = parsed_args.format
        wait = not parsed_args.no_wait

        task = ansible.run.delay("netbox-local", "init", arguments)
        rc = 0

        if wait:
            logger.info(f"Task {task.task_id} was prepared for execution.")
            logger.info(
                f"It takes a moment until task {task.task_id} has been started and output is visible here."
            )
            rc = handle_task(task, wait, format, 300)
        else:
            logger.info(
                f"Task {task.task_id} is running in background. No more output. Check ARA for logs."
            )

        return rc


class Import(Command):
    def get_parser(self, prog_name):
        parser = super(Import, self).get_parser(prog_name)
        parser.add_argument(
            "--vendors",
            help="Vendors from which all available device types are to be imported",
            required=False,
        )
        parser.add_argument(
            "--library",
            default=False,
            help="Do import device types from the device type library",
            action="store_true",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the role has been applied",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        vendors = parsed_args.vendors
        wait = not parsed_args.no_wait

        task = netbox.import_device_types.delay(vendors, parsed_args.library)

        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Manage(Command):
    def get_parser(self, prog_name):
        parser = super(Manage, self).get_parser(prog_name)
        parser.add_argument(
            "--type",
            type=str,
            help="Type of the resource to manage",
            required=False,
            default="rack",
        )
        parser.add_argument(
            "name", nargs="?", type=str, help="Name of the resource to manage"
        )
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Other arguments for Ansible"
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the changes have been made",
            action="store_true",
        )
        parser.add_argument(
            "--format",
            default="log",
            help="Output type",
            const="log",
            nargs="?",
            choices=["script", "log"],
        ),
        return parser

    def take_action(self, parsed_args):
        arguments = parsed_args.arguments
        format = parsed_args.format
        name = parsed_args.name
        type_of_resource = parsed_args.type
        wait = not parsed_args.no_wait

        rc = 0
        if not name:
            logger.info("No name of an object to be managed was given")
            table = []
            for playbook_type in ["rack"]:
                playbooks = glob(
                    f"/opt/configuration/netbox/playbooks/{playbook_type}-*.yml"
                )
                for playbook in playbooks:
                    name = basename(playbook)[len(playbook_type) + 1 : -4]  # noqa E203
                    table.append([playbook_type, name])

            print(tabulate(table, headers=["Type", "Name"], tablefmt="psql"))
        else:
            task = ansible.run.delay(
                "netbox-local", f"{type_of_resource}-{name}", arguments
            )

            if wait:
                logger.info(f"Task {task.task_id} was prepared for execution.")
                logger.info(
                    f"It takes a moment until task {task.task_id} has been started and output is visible here."
                )
                rc = handle_task(task, wait, format, 300)
            else:
                logger.info(
                    f"Task {task.task_id} is running in background. No more output. Check ARA for logs."
                )

        return rc


class Connect(Command):
    def get_parser(self, prog_name):
        parser = super(Connect, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            nargs="?",
            type=str,
            help="Name of the resource (a collection or a device) to connect",
        )
        parser.add_argument(
            "--collection",
            type=str,
            help="Name of the collection to connect",
            required=False,
        ),
        parser.add_argument(
            "--device", type=str, help="Name of the device to connect", required=False
        )
        parser.add_argument(
            "--enforce",
            default=False,
            help="Ignore the current transition of a device",
            action="store_true",
        )
        parser.add_argument(
            "--state", type=str, help="State to use", default=None, required=False
        )
        parser.add_argument(
            "--type",
            type=str,
            default="collection",
            help="Type of the resource to connection (when not using --collection or --device)",
            required=False,
        )
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name
        collection = parsed_args.device
        device = parsed_args.device
        state = parsed_args.state
        type_of_resource = parsed_args.type
        enforce = parsed_args.enforce

        task = None

        if name:
            if type_of_resource == "collection":
                task = netbox.data.delay(name, "", state)
            elif type_of_resource == "device":
                task = netbox.data.delay("", name, state)
        else:
            task = netbox.data.delay(collection, device, state)
            name = f"{collection}-{device}"

        task.wait(timeout=None, interval=0.5)
        data = task.get()

        for device in data:
            t = netbox.connect.delay(device, state, data, enforce)
            logger.info(
                f"Task {t.task_id} for device {device} is running in background"
            )
        logger.info(
            "Tasks are running in background. No more output. Check Flower for logs."
        )


class Disable(Command):
    def get_parser(self, prog_name):
        parser = super(Disable, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            nargs=1,
            type=str,
            help="Name of the device to check for unused interfaces",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the changes have been made",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        wait = not parsed_args.no_wait

        task = netbox.disable.delay(name)

        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Generate(Command):
    def get_parser(self, prog_name):
        parser = super(Generate, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            nargs=1,
            type=str,
            help="Name of the device for which the configuration is to be generated.",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the changes have been made",
            action="store_true",
        )
        parser.add_argument(
            "--template", type=str, help="Name of the template to use", required=False
        )
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        template = parsed_args.template
        wait = not parsed_args.no_wait

        task = netbox.generate.delay(name, template)

        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Deploy(Command):
    def get_parser(self, prog_name):
        parser = super(Deploy, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            nargs=1,
            type=str,
            help="Name of the device for which the configuration is to be deployed",
        )
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Other arguments for Ansible"
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the changes have been made",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        # arguments = parsed_args.arguments
        wait = not parsed_args.no_wait

        task = netbox.deploy.delay(name)

        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Check(Command):
    def get_parser(self, prog_name):
        parser = super(Check, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            nargs=1,
            type=str,
            help="Name of the device for which the configuration is to be checked",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the changes have been made",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        wait = not parsed_args.no_wait

        task = netbox.check.delay(name)

        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Diff(Command):
    def get_parser(self, prog_name):
        parser = super(Diff, self).get_parser(prog_name)
        parser.add_argument(
            "name",
            nargs=1,
            type=str,
            help="Name of the device for which the configuration is to be diffed",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the changes have been made",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        name = parsed_args.name[0]
        wait = not parsed_args.no_wait

        task = netbox.diff.delay(name)

        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Ping(Command):
    def get_parser(self, prog_name):
        parser = super(Ping, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        task = netbox.ping.delay()
        task.wait(timeout=None, interval=0.5)
        result = task.get()
        print(result)
